# Modifications Copyright (C) 2020 Ilias Giechaskiel, https://ilias.giechaskiel.com
# Code adapted from https://github.com/aws/aws-fpga/blob/master/sdk/apps/byte_swapper/fpga_funcs.py

# Copyright 2015-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may
# not use this file except in compliance with the License. A copy of the
# License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# I. Giechaskiel, K. B. Rasmussen, and J. Szefer, "Reading Between the Dies: Cross-SLR Covert
# Channels on Multi-Tenant Cloud FPGAs," IEEE International Conference on Computer Design (ICCD),
# Abu Dhabi, United Arab Emirates, November 2019.

import time

from ctypes import *

import fpga_mgmt
import fpga_pci
import fpga_dma

class FPGA(object):
    def __init__(self, slot, agfi, logger=None):
        self.slot = slot
        self.agfi = agfi
        self.handle = None
        self.logger = logger
        self.log('info', 'AGFI {} to be loaded to slot {}'.format(self.agfi, self.slot))
        fpga_mgmt.fpga_mgmt_init()

    def log(self, level, msg):
        if self.logger:
            getattr(self.logger, level)(msg)

    def raise_error(self, msg):
        self.log('error', msg)
        raise RuntimeError(msg)

    def get_slot_info(self):
        self.log('info', 'Retrieving slot {} info'.format(self.slot))
        ret = c_int32()
        info = fpga_mgmt.struct_fpga_mgmt_image_info()
        info.slot_id = self.slot
        ret = fpga_mgmt.fpga_mgmt_describe_local_image(info.slot_id, byref(info), c_uint32(0))
        if ret != 0:
            self.raise_error('Unable to describe current image on slot {} (returned {})'.format(self.slot, ret))
        return info

    def load_agfi(self):
        info = self.get_slot_info()
        ret = c_int32()
        self.log('info', 'Current AGFI: {}'.format(info.ids.afi_id))
        load_opt = fpga_mgmt.union_fpga_mgmt_load_local_image_options()
        fpga_mgmt.fpga_mgmt_init_load_local_image_options(byref(load_opt))
        load_opt._1.slot_id = self.slot
        load_opt._1.afi_id = cast(self.agfi, POINTER(c_char))
        load_opt._1.flags = fpga_mgmt.FPGA_CMD_FORCE_SHELL_RELOAD
        info = fpga_mgmt.struct_fpga_mgmt_image_info()
        # Choosing arbitrary delay and polling intervals - 3000 and 20 msec
        ret =  fpga_mgmt.fpga_mgmt_load_local_image_sync_with_options(byref(load_opt), 3000, 20, byref(info))
        if ret != 0:
            self.raise_error('Could not load AGFI {}'.format(self.agfi))
        self.log('info', 'Loaded AGFI: {}'.format(info.ids.afi_id))

    def clear_fpga(self):
        self.log('info', 'Clearing slot {}'.format(self.slot))
        ret = c_int32()
        ret = fpga_mgmt.fpga_mgmt_clear_local_image(self.slot)
        if ret != 0:
            self.raise_error('Could not clear slot {} (returned {})'.format(self.slot, ret))
        self.log('info', 'Cleared slot {}'.format(self.slot))

    def clean_up(self, sleep=10):
        self.log('info', 'Cleaning up')
        fpga_mgmt.fpga_mgmt_close()
        time.sleep(sleep)
        self.log('info', 'Cleaned up')

    def get_led_status(self):
        self.log('debug', 'Getting LED values on slot {}'.format(self.slot))
        ret = c_int32()
        ledval = fpga_mgmt.uint16_t()
        ret = fpga_mgmt.fpga_mgmt_get_vLED_status(self.slot, byref(ledval))
        if ret != 0:
            self.raise_error('Could not get LED values on slot {} (returned {})'.format(self.slot, ret))
        return ledval.value

    def set_pci_handle(self):
        if self.handle:
            self.log('debug', 'PCIe handle on slot {} already set'.format(self.slot))
            return
        self.log('info', 'Getting PCIe handle on slot {}'.format(self.slot))
        ret = c_int32()
        handle = fpga_pci.pci_bar_handle_t()
        ret = fpga_pci.fpga_pci_attach(self.slot, fpga_pci.FPGA_APP_PF, fpga_pci.APP_PF_BAR0, 0, byref(handle))
        if ret != 0:
            self.raise_error('Could not attach to slot {} (returned {})'.format(self.slot, ret))
        self.handle = handle

    def peek(self, addr, sleep=0.0001):
        self.log('debug', 'Peeking address {} on slot {}'.format(addr, self.slot))
        self.set_pci_handle()
        ret = c_int32()
        val = fpga_pci.uint32_t()
        time.sleep(sleep)
        ret = fpga_pci.fpga_pci_peek(self.handle, addr, byref(val))
        self.log('debug', 'Peek of address {} on slot {} returned 0x{:08x}'.format(addr, self.slot, val.value))
        if ret != 0:
            self.raise_error('Could not peek from slot {}, address {} (returned {})'.format(self.slot, addr, val))
        return val.value

    def poke(self, addr, val, sleep=0.01):
        self.log('debug', 'Poking value {} into address {} on slot {}'.format(val, addr, self.slot))
        self.set_pci_handle()
        ret = c_int32()
        ret = fpga_pci.fpga_pci_poke(self.handle, addr, val)
        time.sleep(sleep)
        if ret != 0:
            self.raise_error('Could not poke to slot {}, address {}, value {} (returned {})'.format(self.slot, addr, val, ret))
