# Copyright (C) 2020 Ilias Giechaskiel, https://ilias.giechaskiel.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# I. Giechaskiel, K. B. Rasmussen, and J. Szefer, "Reading Between the Dies: Cross-SLR Covert
# Channels on Multi-Tenant Cloud FPGAs," IEEE International Conference on Computer Design (ICCD),
# Abu Dhabi, United Arab Emirates, November 2019.

import argparse
import json
import logging
import os
import time

from aws_utilities import FPGA
from data_parser import DataParser

# Relevant PCIE addresses
class Addresses(object):
    FIFO_ADDR           = 0x0500 # Read-only
    CONFIG_ADDR         = 0x0504 # Read-write
    TRANS_MASK_ADDR     = 0x0510 # Read-write
    STATUS_ADDR         = 0x0514 # Read-only
    SETUP_ADDR          = 0x0518 # Read-only
    STAGES_ADDR         = 0x051c # Read-only
    PATTERN_ADDR        = 0x0520 # Read-write

# Covert channel receiver and transmitter
class CovertChannel(object):
    def __init__(self, fpga):
        self.fpga = fpga

    # Read 32 bits from FIFO (blocks until data is available)
    def read(self):
        while True:
            val = self.fpga.peek(Addresses.STATUS_ADDR)
            self.fpga.log('debug', 'Status is 0x{:08x}'.format(val))
            if (val >> 7) & 1: # Don't yet have value. Sleep
                time.sleep(0.000001)
            else:
                return self.fpga.peek(Addresses.FIFO_ADDR)

    # Set transmitter mask (which transmitters to enable)
    def set_mask(self, val):
        self.fpga.poke(Addresses.TRANS_MASK_ADDR, val)
        self.fpga.log('info', 'Set transmitter mask to 0x{:08x}'.format(val))

    # Set 32-bit transmission pattern (ideally Manchester-encoded)
    def set_pattern(self, val):
        self.fpga.poke(Addresses.PATTERN_ADDR, val)
        self.fpga.log('info', 'Set pattern to 0x{:08x}'.format(val))

    # Set measurement configuration
    def configure(self, params):
        list_vals = [
            (params['trigbit'], 1), # New measurement every 2**t clock cycles
            (params['waitcnt'], 1), # Pipelining-related value
            (params['measurements'], 2), # Number of measurements
        ]

        val = 0
        for v, b in list_vals:
            val = (val << (8*b)) | v
        self.fpga.poke(Addresses.CONFIG_ADDR, val)
        self.fpga.log('info', 'Configured receiver with 0x{:08x}'.format(val))

    # Read FPGA configuration
    def read_config(self):
        # Receiver and transmitter numbers, types, and ROs
        raw_config = self.fpga.peek(Addresses.SETUP_ADDR)
        config = self.parse_config(raw_config)
        self.fpga.log('debug', 'Received configuration 0x{:08x}. Interpreted as {}'.format(raw_config, config))

        # Receiver and transmitter stages
        raw_config = self.fpga.peek(Addresses.STAGES_ADDR)
        stages_config = self.parse_stages(raw_config)
        self.fpga.log('debug', 'Received stages configuration 0x{:08x}. Interpreted as {}'.format(raw_config, stages_config))

        for prefix in ['recv', 'trans']:
            config['{}_stages'.format(prefix)] = stages_config['{}_stages'.format(prefix)]
            # Number of ROs split over the two configurations
            config['num_{}_ros'.format(prefix)] |= stages_config['{}_ros'.format(prefix)]
        self.fpga.log('info', 'Received FPGA configuration: {}'.format(config))
        return config

    # Parse FPGA configuration
    @staticmethod
    def parse_config(raw_config):
        def parse_val(val, prefix):
            desc = [
                (4, 'num_{}'.format(prefix)),       # Number of regions
                (2, '{}_ro_type'.format(prefix)),   # RO type: 0=LUT, 1=Latch, 2=Register
                (10, 'num_{}_ros'.format(prefix)),  # Number of ROs in each region
            ]
            parsed = {}
            for d in desc:
                b, k = d
                v = val & ((1 << b) - 1)
                val >>= b
                parsed[k] = v
            return parsed

        config = {}
        for prefix in ['recv', 'trans']:
            parsed = parse_val(raw_config, prefix)
            raw_config >>= 16
            for k, v in parsed.iteritems():
                config[k] = v
        return config

    # Parse FPGA stage information
    @staticmethod
    def parse_stages(raw_stages):
        def parse_val(val, prefix):
            desc = [
                (8, '{}_stages'.format(prefix)),    # Number of RO intermediate stages
                (8, '{}_ros'.format(prefix)),       # Number of ROs (top bits)
            ]
            parsed = {}
            for d in desc:
                b, k = d
                v = val & ((1 << b) - 1)
                val >>= b
                if k.endswith('ros'):
                    v = v << 10
                parsed[k] = v
            return parsed

        config = {}
        for prefix in ['recv', 'trans']:
            parsed = parse_val(raw_stages, prefix)
            raw_stages >>= 16
            for k, v in parsed.iteritems():
                config[k] = v
        return config

    # Create communication channel and measure results
    def measure(self, params):
        # Receive configuration from FPGA
        config = self.read_config()

        # Update it with experimental parameters
        config['measurements'] = params['measurements'] & 0xFFFF
        config['total_measurements'] = 32*config['measurements']
        config['mask'] = params['mask'] & ((1 << config['num_trans']) - 1)

        outdir = params['outdir']
        outfile = os.path.join(outdir, 'parameters.json')

        if os.path.exists(outfile):
            if params['overwrite']:
                self.fpga.log('warn', 'File {} exists. Overwriting.'.format(outfile))
            else:
                self.fpga.log('warn', 'File {} exists. NOT overwriting.'.format(outfile))
                return

        # Dump configuration and experimental parameters
        json_data = dict(params)
        for k, v in config.iteritems():
            json_data[k] = v
        with open(outfile, 'w') as jsonfile:
            json.dump(json_data, jsonfile, indent=4, sort_keys=True)

        # Set transmitter mask and pattern
        self.set_mask(config['mask'])
        self.set_pattern(params['pattern'])

        # Start measurements
        parser = DataParser(self.read, config)
        repetitions = params['repetitions']
        for i in range(repetitions):
            self.fpga.log('info', 'Repetition {}/{}'.format(i+1, repetitions))

            # Setup experiment and save measurements
            self.configure(params)
            measurements = parser.measure()
            outfile = os.path.join(outdir, '{}.csv'.format(i))
            DataParser.write_measurements(measurements, config['mask'], outfile)

        # Turn all transmitters off
        self.set_mask(0)

if __name__ == '__main__':
    def auto_int(s):
        return int(s, 0)

    parser = argparse.ArgumentParser(description='Cross-SLR communication on AWS.')
    parser.add_argument('-s', '--slot', default=0, type=int, help='FPGA slot')
    parser.add_argument('-a', '--agfi', required=True, help='Amazon Global FPGA Image')
    parser.add_argument('-o', '--outdir', help='Output directory')
    parser.add_argument('-t', '--trigbit', default=7, type=int, help='Trigger bit (2**t clock cycles per measurement)')
    parser.add_argument('-p', '--pattern', default=0xaaaaaaaa, type=auto_int, help='Transmission Pattern') # Default is alternating pattern
    parser.add_argument('-m', '--measurements', default=64, type=int, help='Number of measurements for the whole pattern')
    parser.add_argument('-r', '--repetitions', default=5, type=int, help='Number of repetitions of all measurements')
    parser.add_argument('-k', '--mask', default=0xffff, type=auto_int, help='Transmitter mask')
    parser.add_argument('-l', '--loglevel', default='INFO', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], help='Logging level')
    parser.add_argument('-w', '--overwrite', action='store_true', help='Overwrite measurements')
    args = parser.parse_args()

    # Setup logging
    level = getattr(logging, args.loglevel)
    logger = logging.getLogger("cross_slr_log")
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter('%(asctime)s %(name)s %(levelname)s: %(message)s'))
    logger.addHandler(handler)

    # Setup output directory
    if args.outdir is None:
        args.outdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'cross_slr_{}_{}'.format(args.agfi, args.slot))
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    logger.debug('Output directory set to {}'.format(args.outdir))

    # Setup image
    fpga = FPGA(args.slot, args.agfi, logger)
    fpga.load_agfi()

    # Create covert channel
    parameters = {
        'waitcnt': 5,
        'measurements': args.measurements,
        'repetitions': args.repetitions,
        'mask': args.mask,
        'pattern': args.pattern,
        'trigbit': args.trigbit,
        'outdir': args.outdir,
        'overwrite': args.overwrite,
        'agfi': args.agfi,
        'slot': args.slot,
    }
    channel = CovertChannel(fpga)
    channel.measure(parameters)

    # Clean up
    fpga.clear_fpga()
    fpga.clean_up()
