# Copyright (C) 2020 Ilias Giechaskiel, https://ilias.giechaskiel.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# I. Giechaskiel, K. B. Rasmussen, and J. Szefer, "Reading Between the Dies: Cross-SLR Covert
# Channels on Multi-Tenant Cloud FPGAs," IEEE International Conference on Computer Design (ICCD),
# Abu Dhabi, United Arab Emirates, November 2019.

import argparse
import csv
import json
import math
import os

# Helper method to get length, mean, and std deviation of array
def get_stats(ar):
    num  = len(ar)
    if num:
        mean = float(sum(ar))/num
        var  = float(sum([(x-mean)*(x-mean) for x in ar]))/num
        std  = math.sqrt(var)
    else:
        mean = None
        std  = None
    return num, mean, std

# Parse all CSV outputs in given directory
class DirectoryParser(object):
    def __init__(self, indir):
        with open(os.path.join(args.indir, 'parameters.json')) as jsonfile:
            self.parameters = json.load(jsonfile)

    # Parse directory
    def parse(self, mhz):
        # Clock speed is in MHz.
        # Multiplier calculates the absolute delay in picoseconds,
        # as effectively doing 10**12/10**6
        mult = ((10**6)*(2.**self.parameters['trigbit']))/(2*mhz)

        summary = {
            'errors': 0,
            'total': 0,
        }
        raw = {
            'ons': None,
            'offs': None,
        }

        # For each repetition, parse CSV file and update results
        for i in range(self.parameters['repetitions']):
            parser = CSVParser(os.path.join(args.indir, '{}.csv'.format(i)), self.parameters)
            cur_results = parser.parse()
            for k in ['errors', 'total']:
                summary[k] += cur_results[k]
            for k in ['ons', 'offs']:
                if raw[k] is None:
                    raw[k] = cur_results[k]
                else:
                    for k2 in cur_results[k]:
                        raw[k][k2].extend(cur_results[k][k2])

        # For each of on/off and each RO, save stats
        keys = sorted(raw)
        for index, k in enumerate(keys):
            for k2 in raw[k]:
                num, mean, std = get_stats(raw[k][k2])
                if k2 not in summary:
                    summary[k2] = {}
                else:
                    other_mean = summary[k2]['{}_mean'.format(keys[1-index])]
                    if other_mean is not None and mean is not None:
                        diff = other_mean - mean
                        delay = (mult*diff)/(other_mean*mean)
                    else:
                        diff = None
                        delay = None
                    summary[k2]['diff'] = diff
                    summary[k2]['delay_ps'] = delay
                summary[k2]['{}_num'.format(k)]  = num
                summary[k2]['{}_mean'.format(k)] = mean
                summary[k2]['{}_std'.format(k)]  = std
        return summary

# Parse single CSV output
class CSVParser(object):
    def __init__(self, infile, parameters):
        self.infile = infile
        self.parameters = dict(parameters)

    # Parse file
    def parse(self):
        csvfile = open(self.infile)
        reader = csv.DictReader(csvfile)

        # Setup measurement dictionaries for all ROs
        offs = {}
        ons = {}
        for i in range(self.parameters['num_recv']):
            for j in range(self.parameters['num_recv_ros']):
                k = 'recv_{}_ro_{}'.format(i, j)
                offs[k] = []
                ons[k] = []

        errors = 0
        total = 0
        has_warned = False
        for index, row in enumerate(reader):
            row = {k: int(v, 0) for k, v in row.iteritems()}

            # Sanity-check file
            if row['repetition'] != index:
                raise ValueError('Expected row index to be {} instead of {}'.format(index, row['repetition']))
            cur_bit = (self.parameters['pattern'] >> (31 - (index % 32))) & 1
            expected_mask = self.parameters['mask'] if cur_bit else 0
            if expected_mask != row['trans_enabled']:
                raise ValueError('Expected row mask for index {} to be {} instead of {}'.format(index, expected_mask, row['trans_enabled']))

            # Append to correct array
            if cur_bit:
                d = ons
            else:
                d = offs
            for k in d:
                d[k].append(row[k])

            if index % 2 == 0:
                prev_vals = row
                prev_bit = cur_bit
            else:
                if prev_bit != cur_bit:
                    # If Manchester-encoded, check for errors per-RO
                    for k in d:
                        diff = row[k] - prev_vals[k]
                        if cur_bit and diff >= 0:
                            errors += 1
                        elif not cur_bit and diff <= 0:
                            errors += 1
                        total += 1
                elif not has_warned:
                    print('WARNING: pattern {:08x} in {} is not Manchester-encoded'.format(self.parameters['pattern'], self.infile))
                    has_warned = True

        # Final sanity check
        if index != self.parameters['total_measurements'] - 1:
            raise ValueError('Expected {} measurements instead of {}'.format(self.parameters['total_measurements'], index + 1))

        csvfile.close()
        results = {
            'errors': errors,
            'total': total,
            'ons': ons,
            'offs': offs,
        }
        return results

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Parse cross-SLR CSV measurements.')
    parser.add_argument('-i', '--indir', required=True, help='Input directory')
    parser.add_argument('-c', '--clk', default=125, type=int, help='FPGA clock frequency (MHz)')

    args = parser.parse_args()
    parser = DirectoryParser(args.indir)
    results = parser.parse(args.clk)
    with open(os.path.join(args.indir, 'summary.json'), 'w') as jsonfile:
        json.dump(results, jsonfile, indent=4, sort_keys=True)
        if results['errors'] is not None:
            print("{} errors in {} Manchester-encoded transmissions: {:.2f}% accuracy".format(results['errors'], results['total'], 100-results['errors']*100./results['total']))
