# Copyright (C) 2020 Ilias Giechaskiel, https://ilias.giechaskiel.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# I. Giechaskiel, K. B. Rasmussen, and J. Szefer, "Reading Between the Dies: Cross-SLR Covert
# Channels on Multi-Tenant Cloud FPGAs," IEEE International Conference on Computer Design (ICCD),
# Abu Dhabi, United Arab Emirates, November 2019.

import csv
import sys

# Parse and interpret FPGA data
class DataParser(object):
    def __init__(self, read_func, config):
        self.read = read_func
        self.config = dict(config)

    # Receive and parse measurements
    def measure(self):
        measurements = dict(self.config)
        start_index = None
        for rep in range(measurements['total_measurements']):
            cur_counts = self.get_all_recv_counts()
            cur_measurements = {
                'trans_enabled': cur_counts['trans_enabled'],
                'mask': measurements['mask'],
            }
            for recv in range(measurements['num_recv']):
                # Get measurements for this receiver
                recv_key = self.get_recv_key(recv)

                # Verify the index is correct
                cur_recv = cur_counts[recv_key]
                cur_index = cur_recv['cur_index']
                if start_index is None:
                    start_index = cur_index
                elif cur_index != (start_index + rep) & 0xFFFF:  # Wraps around
                    raise ValueError('Invalid index for receiver {} on repetition {}: Expected {} but got {}'.format(recv, rep, start_index + rep, cur_index))

                # Copy the results for each RO
                cur_recv_measurements = {}
                for ro in range(measurements['num_recv_ros']):
                    ro_key = self.get_ro_key(ro)
                    cur_recv_measurements[ro_key] = cur_recv[ro_key]
                cur_measurements[recv_key] = cur_recv_measurements
            measurements[rep] = cur_measurements
        return measurements

    # Receive single measurement
    # Data format for each measurement:
    #   First 32-bit word: 1...1 (repeated 32-#trans regions times) trans_enabled
    #   Then for each receiver region:
    #     First 32-bit word: 1...1 (11 times), receiver ID (5 bits, 0-#recv regions -1), measurement index (16 bits)
    #     32-bit count for each RO in region
    # I.e., each measurement corresponds to 1+C*(1+R) words (C=5,R=5 =>32*31=992 bit)
    def get_all_recv_counts(self):
        results = {}
        results['trans_enabled'] = self.get_enabled_trans() # Get enabled transmitters
        num_recv = self.config['num_recv']
        for recv in range(num_recv): # For each receiver region
            counts = self.get_single_recv_counts() # Get counts
            # And check the receiver ID
            recv_index = num_recv-recv-1
            recv_id = counts['recv_id']
            if recv_id != recv_index:
                raise ValueError('Invalid Receiver ID. Expected {} but got {}'.format(recv_index, recv_id))
            results[self.get_recv_key(recv_index)] = counts
        return results

    # Get transmitter status
    def get_enabled_trans(self):
        val = self.read()
        orig_val = val

        trans_val = None
        mask = self.config['mask']
        orig_mask = mask
        num_trans = self.config['num_trans']
        # Verify the correct bits are enabled
        for i in range(num_trans):
            cur_trans = val & 1
            val >>= 1
            cur_mask = mask & 1
            mask >>= 1

            # If cur_mask does not enable this transmitter, its value must be 0
            if not cur_mask:
                if cur_trans != 0:
                    raise ValueError('Expected transmitter {} to be off because of mask 0x{:08x}. Instead got 0x{:08x}'.format(i, orig_mask, orig_val))
                continue

            # Otherwise, all enabled transmitters must have the same value
            if trans_val is None:
                trans_val = cur_trans
            else:
                if trans_val != cur_trans:
                    raise ValueError('Expected transmitter {} to have value {} instead of {} for consistency. Instead got 0x{:08x}'.format(i, trans_val, cur_trans, orig_val))

        # MSB bits must be set
        for i in range(32-num_trans):
            if (val & 1) != 1:
                raise ValueError('Expected {} MSB 1s, but got 0x{:08x}'.format(32-num_trans, orig_val))
            val >>= 1

        if trans_val is None or trans_val == 0:
            return 0
        else:
            return ((1 << num_trans) - 1) & orig_mask

    # Get the counts for all ROs of a single receiver region
    def get_single_recv_counts(self):
        val = self.read()
        orig_val = val

        # Verify status word
        results = {}
        results['cur_index'] = val & 0xFFFF
        val >>= 16
        results['recv_id'] = val & 0x1F
        val >>= 5

        for i in range(11):
            if val & 1 != 1:
                raise ValueError('Invalid input 0x{:08x}. Expected eleven 1s as MSBs'.format(orig_val))
            val >>= 1

        # Get individual RO counts
        num_ros = self.config['num_recv_ros']
        for ro in range(num_ros):
            results[self.get_ro_key(num_ros-ro-1)] = self.read()
        return results

    # Identifier for a given receiver region
    @staticmethod
    def get_recv_key(recv_index):
        return 'recv_{}'.format(recv_index)

    # Identifier got a given RO
    @staticmethod
    def get_ro_key(ro_index):
        return 'ro_{}'.format(ro_index)

    @staticmethod
    def write_measurements(measurements, mask, outfile):
        # Setup CSV field names
        fieldnames = ['repetition', 'trans_enabled']
        num_recv = measurements['num_recv']
        num_ros = measurements['num_recv_ros']
        for recv in range(num_recv):
            recv_key = DataParser.get_recv_key(recv)
            for ro in range(num_ros):
                ro_key = DataParser.get_ro_key(ro)
                fieldnames.append('{}_{}'.format(recv_key, ro_key))

        # Write CSV header
        if not outfile:
            csvfile = sys.stdout
        else:
            csvfile = open(outfile, 'wb')
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        # For each measurement
        for rep in range(measurements['total_measurements']):
            cur_meas = measurements[rep]
            line = {
                'trans_enabled': cur_meas['trans_enabled'],
                'repetition': rep
            }
            # For each receiver region
            for recv in range(num_recv):
                recv_key = DataParser.get_recv_key(recv)
                # For each RO
                for ro in range(num_ros):
                    ro_key = DataParser.get_ro_key(ro)
                    key = '{}_{}'.format(recv_key, ro_key)
                    # Save RO count
                    line[key] = cur_meas[recv_key][ro_key]
            writer.writerow(line)

        if outfile:
            csvfile.close()
